/*
 * Copyright 2021 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.tomee.vestige.launcher;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.apache.tomcat.vestige.launcher.VestigeWar;
import org.apache.tomcat.vwar.Application;
import org.apache.tomcat.vwar.Config;
import org.apache.tomcat.vwar.Launcher;
import org.apache.tomcat.vwar.MavenClassType;
import org.apache.tomcat.vwar.MavenConfig;
import org.apache.tomcat.vwar.ObjectFactory;
import org.apache.tomcat.vwar.ReplaceDependency;
import org.apache.tomcat.vwar.VerificationMetadata;
import org.xml.sax.SAXException;

/**
 * @author Gael Lalire
 */
public final class VWarGenerator {

    private VWarGenerator() {
    }

    public static void generate(final WarInEar warInEar, final File destFile) throws IOException, JAXBException, SAXException {
        Application application = new Application();
        application.setJavaSpecificationVersion(warInEar.getJavaVersion());
        Launcher launcher = new Launcher();
        application.setLauncher(launcher);
        MavenClassType mavenResolver = new MavenClassType();
        launcher.setMavenResolver(mavenResolver);
        mavenResolver.setGroupId(warInEar.getGroupId());
        mavenResolver.setArtifactId(warInEar.getArtifactId());
        mavenResolver.setVersion(warInEar.getVersion());
        mavenResolver.setClassifier(warInEar.getClassifier());
        WarVerificationMetadata earVerificationMetadata = warInEar.getVerificationMetadata();
        if (earVerificationMetadata != null) {
            VerificationMetadata verificationMetadata = new VerificationMetadata();
            verificationMetadata.setText(earVerificationMetadata.getText());
            launcher.setVerificationMetadata(verificationMetadata);
        }
        List<WarLibExclusion> warLibExclusions = warInEar.getWarLibExclusions();
        if (warLibExclusions != null) {
            Config config = new Config();
            application.setConfigurations(config);
            MavenConfig mavenConfig = new MavenConfig();
            for (WarLibExclusion warLibExclusion : warLibExclusions) {
                ReplaceDependency replaceDependency = new ReplaceDependency();
                replaceDependency.setGroupId(warLibExclusion.getGroupId());
                replaceDependency.setArtifactId(warLibExclusion.getArtifactId());
                mavenConfig.getModifyDependencyOrReplaceDependencyOrAdditionalRepository().add(replaceDependency);
            }
            config.setMavenConfig(mavenConfig);
        }

        Thread currentThread = Thread.currentThread();
        ClassLoader contextClassLoader = currentThread.getContextClassLoader();
        currentThread.setContextClassLoader(VWarGenerator.class.getClassLoader());
        Marshaller marshaller;
        try {
            JAXBContext jc = JAXBContext.newInstance(ObjectFactory.class.getPackage().getName());
            marshaller = jc.createMarshaller();

            URL xsdURL = VestigeWar.class.getResource("vwar-1.0.0.xsd");
            SchemaFactory schemaFactory = SchemaFactory.newInstance("http://www.w3.org/2001/XMLSchema");
            Schema schema = schemaFactory.newSchema(xsdURL);
            marshaller.setSchema(schema);
            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
        } finally {
            currentThread.setContextClassLoader(contextClassLoader);
        }

        try (FileOutputStream fos = new FileOutputStream(destFile)) {
            marshaller.marshal(new JAXBElement<Application>(new QName("http://tomcat.apache.org/VWAR", "application"), Application.class, application), fos);
        }
    }

}
